r"""
This module contains usefull preprocessing to manipulate side channel traces..

.. currentmodule:: scalib.preprocessing

.. autosummary::
   :toctree:
   :recursive:
   :nosignatures:

   QuantFitMethod
   Quantizer
"""
__all__ = ["QuantFitMethod", "Quantizer"]

from .quantization import QuantFitMethod, Quantizer
